@@ -1,10 +1,10 @@
 # -*- perl -*-
 
 #
-# $Id: Getopt.pm,v 1.31 1999/08/31 10:03:20 eserte Exp $
+# $Id: Getopt.pm,v 1.34 2000/09/02 00:08:35 eserte Exp $
 # Author: Slaven Rezic
 #
-# Copyright (C) 1997,1998,1999 Slaven Rezic. All rights reserved.
+# Copyright (C) 1997,1998,1999,2000 Slaven Rezic. All rights reserved.
 # This package is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 #
@@ -18,8 +18,12 @@ use strict;
 use vars qw($loadoptions $VERSION $x11_pass_through
 	    $CHECKMARK_OFF $CHECKMARK_ON $DEBUG
 	   );
+use constant OPTNAME  => 0;
+use constant OPTTYPE  => 1;
+use constant DEFVAL   => 2;
+use constant OPTEXTRA => 3;
 
-$VERSION = '0.38';
+$VERSION = '0.41';
 
 $DEBUG = 0;
 $x11_pass_through = 0;
@@ -33,16 +37,21 @@ sub new {
     if (exists $a{'-opttable'}) {
 	$self->{'opttable'} = delete $a{'-opttable'};
 	foreach (@{$self->{'opttable'}}) {
+	    # Convert from new style without hash for extra options for
+	    # internal operation.
+	    # ['opt', '=s', 'defval', 'x' => 'y', 'z' => 'a', ...] into
+	    # ['opt', '=s', 'defval', {'x' => 'y', 'z' => 'a', ...}]
 	    if (ref $_ eq 'ARRAY' and
-		defined $_->[3] and
-		ref $_->[3] ne 'HASH') {
-		my %h = splice @$_, 3;
-		$_->[3] = \%h;
+		defined $_->[OPTEXTRA] and
+		ref $_->[OPTEXTRA] ne 'HASH') {
+		my %h = splice @$_, OPTEXTRA;
+		$_->[OPTEXTRA] = \%h;
 	    }
-	    if (ref $_ eq 'ARRAY' && $_->[0] =~ /\|/) { # handle aliases
-		my($opt, @aliases) = split(/\|/, $_->[0]);
-		$_->[0] = $opt;
-		push(@{$_->[3]{'aliases'}}, @aliases);
+	    # Handle aliases
+	    if (ref $_ eq 'ARRAY' && $_->[OPTNAME] =~ /\|/) {
+		my($opt, @aliases) = split(/\|/, $_->[OPTNAME]);
+		$_->[OPTNAME] = $opt;
+		push(@{$_->[OPTEXTRA]{'aliases'}}, @aliases);
 	    }
 	}
     } elsif (exists $a{'-getopt'}) {
@@ -121,17 +130,20 @@ sub new {
     $self->{'filename'} = delete $a{'-filename'};
     $self->{'nosafe'}   = delete $a{'-nosafe'};
 
-    die "Extra arguments: " . join(" ", %a) if %a;
+    die "Unrecognized arguments: " . join(" ", %a) if %a;
 
     bless $self, $pkg;
 }
 
-# Return a list with all option names
+# Return a list with all option names, that is, section labels and
+# descriptions are ignored.
 sub _opt_array {
     my $self = shift;
     my @res;
     foreach (@{$self->{'opttable'}}) {
-	push @res, $_ if ref $_ eq 'ARRAY';
+	push @res, $_
+	    if ref $_ eq 'ARRAY' and
+	       $_->[OPTNAME] ne '';
     }
     @res;
 }
@@ -139,15 +151,15 @@ sub _opt_array {
 # Return a reference to the option variable given by $opt
 sub _varref {
     my($self, $opt) = @_;
-    if($opt->[3]{'var'}) {
-	$opt->[3]{'var'};
+    if($opt->[OPTEXTRA]{'var'}) {
+	$opt->[OPTEXTRA]{'var'};
     } elsif ($self->{'options'}) {
-	\$self->{'options'}{$opt->[0]};
+	\$self->{'options'}{$opt->[OPTNAME]};
     } else {
 	# Link to global $opt_XXX variable.
 	# Make sure a valid perl identifier results.
 	my $v;
-	($v = $opt->[0]) =~ s/\W/_/g;
+	($v = $opt->[OPTNAME]) =~ s/\W/_/g;
 	eval q{\$} . $self->{'caller'} . q{::opt_} . $v; # XXX @, %
     }
 }
@@ -156,8 +168,8 @@ sub set_defaults {
     my $self = shift;
     my $opt;
     foreach $opt ($self->_opt_array) {
-	if (defined $opt->[2]) {
-	    $ {$self->_varref($opt)} = $opt->[2];
+	if (defined $opt->[DEFVAL]) {
+	    $ {$self->_varref($opt)} = $opt->[DEFVAL];
 	}
     }
 }
@@ -184,17 +196,17 @@ sub load_options {
 
     my $opt;
     foreach $opt ($self->_opt_array) {
-	if (exists $loadoptions->{$opt->[0]}) {
+	if (exists $loadoptions->{$opt->[OPTNAME]}) {
 	    if (ref $self->_varref($opt) eq 'CODE') {
-		&{$self->_varref($opt)} if $loadoptions->{$opt->[0]};
+		&{$self->_varref($opt)} if $loadoptions->{$opt->[OPTNAME]};
 	    } elsif (ref $self->_varref($opt) eq 'ARRAY' &&
-		     ref $loadoptions->{$opt->[0]} eq 'ARRAY') {
-		@{ $self->_varref($opt) } = @{ $loadoptions->{$opt->[0]} };
+		     ref $loadoptions->{$opt->[OPTNAME]} eq 'ARRAY') {
+		@{ $self->_varref($opt) } = @{ $loadoptions->{$opt->[OPTNAME]} };
 	    } elsif (ref $self->_varref($opt) eq 'HASH' &&
-		     ref $loadoptions->{$opt->[0]} eq 'HASH') {
-		%{ $self->_varref($opt) } = %{ $loadoptions->{$opt->[0]} };
+		     ref $loadoptions->{$opt->[OPTNAME]} eq 'HASH') {
+		%{ $self->_varref($opt) } = %{ $loadoptions->{$opt->[OPTNAME]} };
 	    } else {
-		$ {$self->_varref($opt)} = $loadoptions->{$opt->[0]};
+		$ {$self->_varref($opt)} = $loadoptions->{$opt->[OPTNAME]};
 	    }
 	}
     }
@@ -214,11 +226,11 @@ sub save_options {
 	    my %saveoptions;
 	    my $opt;
 	    foreach $opt ($self->_opt_array) {
-		if (!$opt->[3]{'nosave'}) {
+		if (!$opt->[OPTEXTRA]{'nosave'}) {
 		    if (ref($self->_varref($opt)) eq 'SCALAR') {
-			$saveoptions{$opt->[0]} = $ {$self->_varref($opt)}
+			$saveoptions{$opt->[OPTNAME]} = $ {$self->_varref($opt)}
 		    } elsif (ref($self->_varref($opt)) =~ /^(HASH|ARRAY)$/) {
-			$saveoptions{$opt->[0]} = $self->_varref($opt);
+			$saveoptions{$opt->[OPTNAME]} = $self->_varref($opt);
 		    } 
 		}
 	    }
@@ -246,11 +258,11 @@ sub get_options {
     my %getopt;
     my $opt;
     foreach $opt ($self->_opt_array) {
-	$getopt{_getopt_long_string($opt->[0], $opt->[1])} =
+	$getopt{_getopt_long_string($opt->[OPTNAME], $opt->[OPTTYPE])} =
 	  $self->_varref($opt);
 	# process aliases
-	foreach (@{$opt->[3]{'alias'}}) {
-	    $getopt{_getopt_long_string($_, $opt->[1])} =
+	foreach (@{$opt->[OPTEXTRA]{'alias'}}) {
+	    $getopt{_getopt_long_string($_, $opt->[OPTTYPE])} =
 	      $self->_varref($opt);
 	}
     }
@@ -321,11 +333,11 @@ sub usage {
  	$usage .= join(', ',
  		       sort { length $a <=> length $b }
  		       map { _getopt_long_dash($_) }
- 		       map { ($opt->[1] eq '!' ? "[no]" : "") . $_ }
-		       ($opt->[0], @{$opt->[3]{'alias'}}));
+ 		       map { ($opt->[OPTTYPE] eq '!' ? "[no]" : "") . $_ }
+		       ($opt->[OPTNAME], @{$opt->[OPTEXTRA]{'alias'}}));
 	$usage .= "\t";
-	$usage .= $opt->[3]{'help'}                if $opt->[3]{'help'};
-	$usage .= " (default: " . $opt->[2] . ") " if $opt->[2];
+	$usage .= $opt->[OPTEXTRA]{'help'}              if $opt->[OPTEXTRA]{'help'};
+	$usage .= " (default: " . $opt->[DEFVAL] . ") " if $opt->[DEFVAL];
 	$usage .= "\n";
     }
     $usage;
@@ -335,8 +347,8 @@ sub process_options {
     my($self, $former, $fromgui) = @_;
     my $options = $self->{'options'};
     foreach ($self->_opt_array) {
-	my $opt = $_->[0];
-	if ($_->[3]{'callback'}) {
+	my $opt = $_->[OPTNAME];
+	if ($_->[OPTEXTRA]{'callback'}) {
 	    # no warnings here ... it would be too complicated to catch
 	    # all undefined values
 	    my $old_w = $^W;
@@ -346,14 +358,14 @@ sub process_options {
 		  && (!exists $former->{$opt}
 		      || $ {$self->_varref($_)} eq $former->{$opt}))) {
 		local($^W) = $old_w; # fall back to original value
-		&{$_->[3]{'callback'}};
+		&{$_->[OPTEXTRA]{'callback'}};
 	    }
 	}
-	if ($_->[3]{'strict'}) {
+	if ($_->[OPTEXTRA]{'strict'}) {
 	    # check for valid values (valid are: choices and default value)
 	    my $v = $ {$self->_varref($_)};
-	    my @choices = @{$_->[3]{'choices'}};
-	    push(@choices, $_->[2]) if defined $_->[2];
+	    my @choices = @{$_->[OPTEXTRA]{'choices'}};
+	    push(@choices, $_->[DEFVAL]) if defined $_->[DEFVAL];
 	    if (!grep(/^$v$/, @choices)) {
 		if (defined $former) {
 		    warn "Not allowed: " . $ {$self->_varref($_)}
@@ -388,38 +400,42 @@ sub _number_widget {
     my($self, $frame, $opt) = @_;
     $frame->Scale
       (-orient => 'horizontal',
-       -from => $opt->[3]{'range'}[0],
-       -to => $opt->[3]{'range'}[1],
+       -from => $opt->[OPTEXTRA]{'range'}[0],
+       -to => $opt->[OPTEXTRA]{'range'}[1],
        -showvalue => 1,
-       -resolution => ($opt->[1] =~ /f/ ? 0 : 1),
+       -resolution => ($opt->[OPTTYPE] =~ /f/ ? 0 : 1),
        -variable => $self->_varref($opt)
       );
 }
 
 sub _integer_widget {
     my($self, $frame, $opt) = @_;
-    if (exists $opt->[3]{'range'}) {
+    if (exists $opt->[OPTEXTRA]{'range'}) {
 	$self->_number_widget($frame, $opt);
     } else {
-	$self->_string_widget($frame, $opt);
+	$self->_string_widget($frame, $opt, -restrict => "=i");
     }
 }
 
 sub _float_widget {
     my($self, $frame, $opt) = @_;
-    if (exists $opt->[3]{'range'}) {
+    if (exists $opt->[OPTEXTRA]{'range'}) {
 	$self->_number_widget($frame, $opt);
     } else {
-	$self->_string_widget($frame, $opt);
+	$self->_string_widget($frame, $opt, -restrict => "=f");
     }
 }
 
 sub _list_widget {
     my($self, $frame, $opt) = @_;
     require Tk::BrowseEntry;
-    my $w = $frame->BrowseEntry(-variable => $self->_varref($opt));
-    my @optlist = @{$opt->[3]{'choices'}};
-    unshift(@optlist, $opt->[2]) if defined $opt->[2];
+    my %args = (-variable => $self->_varref($opt));
+    if ($opt->[OPTEXTRA]{'strict'}) {
+	$args{-state} = "readonly";
+    }
+    my $w = $frame->BrowseEntry(%args);
+    my @optlist = @{$opt->[OPTEXTRA]{'choices'}};
+    unshift(@optlist, $opt->[DEFVAL]) if defined $opt->[DEFVAL];
     my $o;
     foreach $o (@optlist) {
 	$w->insert("end", $o);
@@ -428,23 +444,45 @@ sub _list_widget {
 }
 
 sub _string_widget {
-    my($self, $frame, $opt) = @_;
-    if (exists $opt->[3]{'choices'}) {
+    my($self, $frame, $opt, %args) = @_;
+    if (exists $opt->[OPTEXTRA]{'choices'}) {
 	$self->_list_widget($frame, $opt);
     } else {
-	$frame->Entry(-textvariable => $self->_varref($opt));
+	my $e = $frame->Entry(-textvariable => $self->_varref($opt));
+	if ($args{-restrict}) {
+	    eval {
+		if ($args{-restrict} eq "=i") {
+		    $e->configure
+			(-validate => "all",
+			 -vcmd => sub {
+			     $_[0] =~ /^([+-]?\d+|)$/;
+			 },
+			);
+		} elsif ($args{-restrict} eq "=f") {
+		    $e->configure
+			(-validate => "all",
+			 -vcmd => sub {
+			     $_[0] =~ /^(|([+-]?)(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))?)$/;
+			 },
+			);
+		}
+	    };
+	    warn $@ if $@;
+	}
+	$e;
     }
 }
 
 sub _filedialog_widget {
-    my($self, $frame, $opt) = @_;
+    my($self, $frame, $opt, %args) = @_;
+    my $subtype = (exists $args{'-subtype'} ? $args{'-subtype'} : 'file');
     my $topframe = $frame->Frame;
     my $e;
-    if (exists $opt->[3]{'choices'}) {
+    if (exists $opt->[OPTEXTRA]{'choices'}) {
 	require Tk::BrowseEntry;
 	$e = $topframe->BrowseEntry(-variable => $self->_varref($opt));
-	my @optlist = @{$opt->[3]{'choices'}};
-	unshift(@optlist, $opt->[2]) if defined $opt->[2];
+	my @optlist = @{$opt->[OPTEXTRA]{'choices'}};
+	unshift(@optlist, $opt->[DEFVAL]) if defined $opt->[DEFVAL];
 	my $o;
 	foreach $o (@optlist) {
 	    $e->insert("end", $o);
@@ -459,11 +497,14 @@ sub _filedialog_widget {
        -command => sub {
 	   require File::Basename;
 	   my($fd, $filedialog);
-	   if ($Tk::VERSION >= 800) {
+	   if ($Tk::VERSION >= 800 && $subtype ne 'dir') {
 	       $fd = 'getOpenFile';
 	   } else {
 	       $fd = 'FileDialog';
-	       eval { require Tk::FileDialog };
+	       eval {
+		   die "nope" if $subtype eq 'dir';
+		   require Tk::FileDialog;
+	       };
 	       if ($@) {
 		   require Tk::FileSelect;
 		   $fd = 'FileSelect';
@@ -494,13 +535,23 @@ sub _filedialog_widget {
 		   $file = $filedialog->Show(-Path => $dir,
 					     -File => $base);
 	       } else {
-		   $file = $filedialog->Show(-directory => $dir);
+		   if ($subtype eq 'dir') {
+		       $file = $filedialog->Show(-directory => $dir,
+						 -verify => [qw(-d)],
+						);
+		   } else {
+		       $file = $filedialog->Show(-directory => $dir);
+		   }
 	       }
 	   } else {
 	       if ($fd eq 'getOpenFile') {
 		   $file = $topframe->getOpenFile(-title => 'Select file');
 	       } else {
-		   $file = $filedialog->Show;
+		   if ($subtype eq 'dir') {
+		       $file = $filedialog->Show(-verify => [qw(-d)]);
+		   } else {
+		       $file = $filedialog->Show;
+		   }
 	       }
 	   }
 	   if ($file) {
@@ -517,19 +568,34 @@ sub _filedialog_widget {
 #   $current_top: title of Notebook page
 #   $optlist: list of options for this Notebook page
 #   $balloon: Balloon widget
+#   $msglist: (optional) list of messages for this Notebook page
 sub _create_page {
-    my($self, $current_page, $optnote, $current_top, $optlist, $balloon) = @_;
+    my($self, $current_page,
+       $optnote, $current_top, $optlist,
+       $balloon, $msglist) = @_;
     $current_page = $optnote->{$current_top} if !defined $current_page;
     my $opt;
     my $row = -1;
+
+    my $msgobj;
+    if (ref $msglist and
+	exists $msglist->{$current_top} and
+	$msglist->{$current_top} ne "") {
+	$row++;
+	$msgobj = $current_page->Label(-text => $msglist->{$current_top},
+				       -justify => "left",
+				       )->grid(-row => $row, -column => 0,
+					       -columnspan => 3);
+    }
+
     foreach $opt (@{$optlist->{$current_top}}) {
 	my $f = $current_page;
 	my $label;
 	my $w;
-	if (exists $opt->[3]{'label'}) {
-	    $label = $opt->[3]{'label'};
+	if (exists $opt->[OPTEXTRA]{'label'}) {
+	    $label = $opt->[OPTEXTRA]{'label'};
 	} else {
-	    $label = $opt->[0];
+	    $label = $opt->[OPTNAME];
 	    if ($label =~ /^(.*)-(.*)$/ && $1 eq $current_top) {
 		$label = $2;
 	    }
@@ -537,43 +603,54 @@ sub _create_page {
 	$row++;
 	my $lw = $f->Label(-text => $label)->grid(-row => $row, -column => 0,
 						  -sticky => 'w');
-	if (exists $opt->[3]{'widget'}) {
+	if (exists $opt->[OPTEXTRA]{'widget'}) {
 	    # own widget
-	    $w = &{$opt->[3]{'widget'}}($self, $f, $opt);
-	} elsif (defined $opt->[1] && $opt->[1] eq '!' or $opt->[1] eq '') {
+	    $w = &{$opt->[OPTEXTRA]{'widget'}}($self, $f, $opt);
+	} elsif (defined $opt->[OPTTYPE] &&
+		 $opt->[OPTTYPE] eq '!' or $opt->[OPTTYPE] eq '') {
 	    $w = $self->_boolean_widget($f, $opt); # XXX _checkmark_
-	} elsif (defined $opt->[1] && $opt->[1] =~ /i/) {
+	} elsif (defined $opt->[OPTTYPE] && $opt->[OPTTYPE] =~ /i/) {
 	    $w = $self->_integer_widget($f, $opt);
-	} elsif (defined $opt->[1] && $opt->[1] =~ /f/) {
+	} elsif (defined $opt->[OPTTYPE] && $opt->[OPTTYPE] =~ /f/) {
 	    $w = $self->_float_widget($f, $opt);
-	} elsif (defined $opt->[1] && $opt->[1] =~ /s/) {
-	    if (defined $opt->[3] && exists $opt->[3]{'subtype'} &&
-		$opt->[3]{'subtype'} eq 'file') {
+	} elsif (defined $opt->[OPTTYPE] && $opt->[OPTTYPE] =~ /s/) {
+	    if (defined $opt->[OPTEXTRA] &&
+		exists $opt->[OPTEXTRA]{'subtype'} &&
+		$opt->[OPTEXTRA]{'subtype'} eq 'file') {
 		$w = $self->_filedialog_widget($f, $opt);
+	    } elsif (defined $opt->[OPTEXTRA] &&
+		     exists $opt->[OPTEXTRA]{'subtype'} &&
+		     $opt->[OPTEXTRA]{'subtype'} eq 'dir') {
+		$w = $self->_filedialog_widget($f, $opt, -subtype => "dir");
 	    } else {
 		$w = $self->_string_widget($f, $opt);
 	    }
 	} else {
-	    warn "Can't generate option editor entry for $opt->[0]";
+	    warn "Can't generate option editor entry for $opt->[OPTNAME]";
 	}
 	if (defined $w) {
 	    $w->grid(-row => $row, -column => 1, -sticky => 'w');
 	}
-	if (exists $opt->[3]{'help'} && defined $balloon) {
-	    $balloon->attach($w, -msg => $opt->[3]{'help'}) if defined $w;
-	    $balloon->attach($lw, -msg => $opt->[3]{'help'}) if defined $lw;
+	if (exists $opt->[OPTEXTRA]{'help'} && defined $balloon) {
+	    $balloon->attach($w, -msg => $opt->[OPTEXTRA]{'help'})
+		if defined $w;
+	    $balloon->attach($lw, -msg => $opt->[OPTEXTRA]{'help'})
+		if defined $lw;
 	}
-	if (exists $opt->[3]{'longhelp'}) {
+	if (exists $opt->[OPTEXTRA]{'longhelp'}) {
 	    $f->Button(-text => '?',
 		       -padx => 1,
 		       -pady => 1,
 		       -command => sub {
-			   my $t = $f->Toplevel(-title => $label);
-			   $t->Label(-text => $opt->[3]{'longhelp'},
+			   my $t = $f->Toplevel
+			       (-title => $self->{_string}{"helpfor"}
+				. " $label");
+			   $t->Label(-text => $opt->[OPTEXTRA]{'longhelp'},
 				     -justify => 'left')->pack;
 			   $t->Button(-text => 'OK',
 				      -command => sub { $t->destroy }
 				     )->pack;
+			   $t->Popup(-popover => "cursor");
 		       })->grid(-row => $row, -column => 2, -sticky => 'w');
 	}
     }
@@ -585,11 +662,11 @@ sub _do_undo {
     my($self, $undo_options) = @_;
     my $opt;
     foreach $opt ($self->_opt_array) {
-	next if $opt->[3]{'nogui'};
-	if (exists $undo_options->{$opt->[0]}) {
+	next if $opt->[OPTEXTRA]{'nogui'};
+	if (exists $undo_options->{$opt->[OPTNAME]}) {
 	    my $swap = $ {$self->_varref($opt)};
-	    $ {$self->_varref($opt)} = $undo_options->{$opt->[0]};
-	    $undo_options->{$opt->[0]} = $swap;
+	    $ {$self->_varref($opt)} = $undo_options->{$opt->[OPTNAME]};
+	    $undo_options->{$opt->[OPTNAME]} = $swap;
 	}
     }
 }
@@ -598,7 +675,9 @@ sub option_editor {
     my($self, $top, %a) = @_;
     my $callback  = delete $a{'-callback'};
     my $nosave    = delete $a{'-nosave'};
+    my $buttons   = delete $a{'-buttons'};
     my $toplevel  = delete $a{'-toplevel'} || 'Toplevel';
+    my $pack      = delete $a{'-pack'};
     my $use_statusbar = delete $a{'-statusbar'};
     my $wait      = delete $a{'-wait'};
     my $string    = delete $a{'-string'};
@@ -610,14 +689,18 @@ sub option_editor {
 		   'defaults'  => 'Defaults',
 		   'ok'        => 'OK',
 		   'apply'     => 'Apply',
-		   'cancel'    => 'Cancel'};
+		   'cancel'    => 'Cancel',
+		   'helpfor'   => 'Help for:',
+	          };
     }
+    $self->{_string} = $string;
+
     # store old values for undo
     my %undo_options;
     my $opt;
     foreach $opt ($self->_opt_array) {
-	next if $opt->[3]{'nogui'};
-	$undo_options{$opt->[0]} = $ {$self->_varref($opt)};
+	next if $opt->[OPTEXTRA]{'nogui'};
+	$undo_options{$opt->[OPTNAME]} = $ {$self->_varref($opt)};
     }
 
     require Tk;
@@ -638,8 +721,9 @@ sub option_editor {
     eval { require Tk::Balloon };
     $dont_use_balloon = 1 if $@;
 
-    my $opt_editor = eval '$top->' . $toplevel . '(%a)';
-    die $@ if $@;
+    my $cmd = '$top->' . $toplevel . '(%a)';
+    my $opt_editor = eval $cmd;
+    die "$@ while evaling $cmd" if $@;
     eval { $opt_editor->configure(-title => $string->{optedit}) };
     my $opt_notebook = ($dont_use_notebook ?
 			$opt_editor->Frame :
@@ -655,25 +739,28 @@ sub option_editor {
     }
 
     my $optlist = {};
+    my $msglist = {};
     my $current_top;
     if ($dont_use_notebook) {
 	$current_top = $string->{'optedit'};
 	foreach $opt ($self->_opt_array) {
 	    push(@{$optlist->{$current_top}}, $opt)
-	      if !$opt->[3]{'nogui'};
+	      if !$opt->[OPTEXTRA]{'nogui'};
 	}
+	# XXX message missing
 	$self->_create_page($opt_notebook, undef, $current_top,
 			    $optlist, $balloon);
     } else {
 	my @opttable = @{$self->{'opttable'}};
 	unshift(@opttable, $string->{'optedit'})
-	  if ref $opttable[0] eq 'ARRAY'; # put head
+	  if ref $opttable[OPTNAME] eq 'ARRAY'; # put head
 	foreach $opt (@opttable) {
 	    if (ref $opt ne 'ARRAY') {
 		my $label = $opt;
 		$current_top = lc($label);
 		my $c = $current_top;
 		$optlist->{$c} = [];
+		$msglist->{$c} = "";
 		$opt_notebook->add($c,
 				   -label => $label,
 				   -anchor => 'w',
@@ -682,11 +769,13 @@ sub option_editor {
 				       $self->_create_page
 					 ($_[0],
 					  $opt_notebook, $c,
-					  $optlist, $balloon);
+					  $optlist, $balloon, $msglist);
                                    });
+            } elsif ($opt->[OPTNAME] eq '') {
+		$msglist->{$current_top} = $opt->[DEFVAL];
 	    } else {
-		push(@{$optlist->{$current_top}}, $opt)
-		  if !$opt->[3]{'nogui'};
+		push @{$optlist->{$current_top}}, $opt
+		  if !$opt->[OPTEXTRA]{'nogui'};
 	    }
 	}
     }
@@ -702,7 +791,7 @@ sub option_editor {
 	   my $nenner = int(($f->Width-2*$bw)/$f->{Sw});
 	   return if (!$nenner);
 	   my $rows = @{$f->{Slaves}}/$nenner;
-	   return if (!$rows);
+	   return if (!$rows or !int($rows));
 	   if ($rows/int($rows) > 0) {
 	       $rows = int($rows)+1;
 	   }
@@ -717,74 +806,107 @@ sub option_editor {
 		 }
 	     });
     my @tiler_b;
-    my $ok_button
-      = $f->Button(-text => $string->{'ok'},
-		   -underline => 0,
-		   -command => sub {
-		       $self->process_options(\%undo_options, 1);
-		       if (!$dont_use_notebook) {
-			   $self->{'raised'} = $opt_notebook->raised();
-		       }
-		       $opt_editor->destroy;
-		   }
-		  );
-    push @tiler_b, $ok_button;
-    my $apply_button
-      = $f->Button(-text => $string->{'apply'},
-		   -command => sub {
-		       $self->process_options(\%undo_options, 1);
-		   }
-		  );
-    push @tiler_b, $apply_button;
-    my $cancel_button
-      = $f->Button(-text => $string->{'cancel'},
-		   -command => sub {
-		       $self->_do_undo(\%undo_options);
-		       if (!$dont_use_notebook) {
-			   $self->{'raised'} = $opt_notebook->raised();
-		       }
-		       $opt_editor->destroy;
-		   }
-		  );
-    push @tiler_b, $cancel_button;
-    my $grid_col = 0;
-    my $undo_button = $f->Button(-text => $string->{'undo'},
-	       -command => sub {
-		   $self->_do_undo(\%undo_options);
-	       }
-	      );
-    push @tiler_b, $undo_button;
-    if ($self->{'filename'}) {
-	my $lastsaved_button = $f->Button(-text => $string->{'lastsaved'},
-		   -command => sub {
-			$top->Busy;
-			$self->load_options;
-			$top->Unbusy;
-		    }
-		  );
-	push @tiler_b, $lastsaved_button;
 
-	if (!$nosave) {
-	    my $sb;
-	    $sb = $f->Button(-text => $string->{'save'},
+    my %allowed_button;
+    if ($buttons) {
+	if (ref $buttons ne 'ARRAY') {
+	    undef $buttons;
+	} else {
+	    %allowed_button = map { ($_ => 1) } @$buttons;
+        }
+    }
+
+    my($ok_button, $apply_button, $cancel_button, $undo_button,
+       $lastsaved_button, $save_button, $def_button);
+
+    if (!$buttons || $allowed_button{'ok'}) {
+	$ok_button
+	    = $f->Button(-text => $string->{'ok'},
+			 -underline => 0,
+			 -command => sub {
+			     $self->process_options(\%undo_options, 1);
+                             if (!$dont_use_notebook) {
+                                 $self->{'raised'} = $opt_notebook->raised();
+                             }
+                             $opt_editor->destroy;
+                         }
+                        );
+        push @tiler_b, $ok_button;
+    }
+
+    if (!$buttons || $allowed_button{'apply'}) {
+	$apply_button
+	    = $f->Button(-text => $string->{'apply'},
+			 -command => sub {
+			     $self->process_options(\%undo_options, 1);
+			 }
+			);
+	push @tiler_b, $apply_button;
+    }
+
+    if (!$buttons || $allowed_button{'cancel'}) {
+	$cancel_button
+	    = $f->Button(-text => $string->{'cancel'},
+			 -command => sub {
+			     $self->_do_undo(\%undo_options);
+			     if (!$dont_use_notebook) {
+				 $self->{'raised'} = $opt_notebook->raised();
+			     }
+			     $opt_editor->destroy;
+			 }
+			);
+	push @tiler_b, $cancel_button;
+    }
+
+    if (!$buttons || $allowed_button{'undo'}) {
+	$undo_button
+	    = $f->Button(-text => $string->{'undo'},
+			 -command => sub {
+			     $self->_do_undo(\%undo_options);
+			 }
+			);
+	push @tiler_b, $undo_button;
+    }
+
+    if ($self->{'filename'}) {
+	if (!$buttons || $allowed_button{'lastsaved'}) {
+	    $lastsaved_button
+		= $f->Button(-text => $string->{'lastsaved'},
+			     -command => sub {
+				 $top->Busy;
+				 $self->load_options;
+				 $top->Unbusy;
+			     }
+			    );
+	    push @tiler_b, $lastsaved_button;
+	}
+
+	if (!$nosave && (!$buttons || $allowed_button{'save'})) {
+	    $save_button
+		= $f->Button(-text => $string->{'save'},
 			     -command => sub {
 				 $top->Busy;
 				 eval { $self->save_options };
 				 if ($@ =~ /No Data::Dumper/) {
-				     $sb->configure(-state => 'disabled');
+				     $save_button->configure(-state => 'disabled');
 				 }
 				 $top->Unbusy;
 			     }
 			    );
-	    push @tiler_b, $sb;
+	    push @tiler_b, $save_button;
 	}
     }
-    my $def_button = $f->Button(-text => $string->{'defaults'},
-				-command => sub {
-				    $self->set_defaults;
-				}
-	      );
-    push @tiler_b, $def_button;
+
+    if (!$buttons || $allowed_button{'defaults'}) {
+	$def_button
+	    = $f->Button(-text => $string->{'defaults'},
+			 -command => sub {
+			     $self->set_defaults;
+			 }
+			);
+	push @tiler_b, $def_button;
+    }
+
     $f->Manage(@tiler_b);
 
     &$callback($self, $opt_editor) if $callback;
@@ -805,8 +927,12 @@ sub option_editor {
 	$opt_editor->Popup;
     }
     if ($wait) {
+	if ($pack) {
+	    $opt_editor->pack(@$pack);
+	}
 	my $wait_var = 1;
 	$opt_editor->OnDestroy(sub { undef $wait_var });
+	$opt_editor->waitVisibility;
 	$opt_editor->grab;
 	$opt_editor->waitVariable(\$wait_var);
     }
@@ -896,7 +1022,7 @@ or B<-opttable> are mandatory.
 
 The arguments for B<new> are:
 
-=over 8
+=over 4
 
 =item -getopt
 
@@ -913,112 +1039,38 @@ Example:
 
 =item -opttable
 
-B<-opttable> provides a more powerful interface. The options are stored
-in variables named I<$opt_XXX> or in a hash when B<-options> is given (see
-below). B<-opttable> should be a reference to an array containing
-all options. Elements of this array may
-be strings, which indicate the beginning of a new group, or array references
-describing the options. The first element of this array is the name of
-the option, the second is the type (C<=s> for string, C<=i> for integer,
+B<-opttable> provides a more powerful interface. The options are
+stored in variables named I<$opt_XXX> or in a hash when B<-options> is
+given (see below). B<-opttable> should be a reference to an array
+containing all options. Elements of this array may be strings, which
+indicate the beginning of a new group, or array references describing
+the options. The first element of this array is the name of the
+option, the second is the type (C<=s> for string, C<=i> for integer,
 C<!> for boolean, C<=f> for float etc., see L<Getopt::Long>) for a
-detailed list. The third element is optional and contains
-the default value (otherwise the default is undefined). Further
-elements are optional too and describes more attributes. The attributes
-have to be key-value pairs with following keys:
+detailed list. The third element is optional and contains the default
+value (otherwise the default is undefined). Further elements are
+optional too and describe more attributes. For a complete list of
+these attributes refer to L<"OPTTABLE ARGUMENTS">.
 
-=over 12
+If an option has no name, then the third element in the description
+array will be used as an global message for the current option page.
+This message can be multi-line.
+Example:
+    ['', '', 'This is an explanation for this option group.']
 
-=item alias
-
-An array of aliases also accepted by F<Getopt::Long>.
-
-=item label
-
-A label to be displayed in the GUI instead of the option name.
-
-=item help
-
-A short help string used by B<usage> and the Balloon help facility in
-B<option_editor>.
-
-=item longhelp
-
-A long help string used by B<option_editor>.
-
-=item choices
-
-An array of additional choices for the option editor.
-
-=item range
-
-An array with the beginning and end of a range for an integer or float value.
-
-=item strict
-
-Must be used with B<choices> or B<range>. When set to true, options have to
-match either the choices or the range.
-
-=item subtype
-
-Current the only permitted subtype is I<file> to be used with string
-options. The GUI interface will pop up a file dialog for this option.
-
-=item var
-
-Use variable instead of I<$options-E<gt>{optname}> or I<$opt_optname>
-to store the value.
-
-=item nogui
-
-This option will not have an entry in the GUI.
-
-=item widget
-
-This should be a reference to a subroutine for creating an own widget.
-Folowing arguments will be passed to this subroutine:
-a reference to the F<Tk::Getopt> object, Frame object, options entry.
-The subroutine should create a widget in the frame (packing is not
-necessary!) and should return a reference to the created widget.
-
-=back
-
-Here is an example for this rather complicated argument:
+Here is an example for a simple opttable:
 
     @opttable =
-        ('Misc',   # Head of first group
-	 ['debug', # name of the option (--debug)
-          '!',     # type boolean, accept --nodebug
-          0,       # default is 0 (false)
-          callback => sub { $^W = 1
-                                if $options->{'debug'}; }
-          # additional attribute: callback to be called if
-          # you set or change the value
-          ],
-         ['age',
-          '=i',    # option accepts integer value
-          18,
-          strict => 1, # value must be in range
-          range => [0, 100], # allowed range
-          alias => ['year', 'years'] # possible aliases
-          ],
-	 'External', # Head of second group
-         ['browser',
-          '=s',    # option accepts string value
-          'tkweb',
-          choices => ['mosaic', 'netscape',
-                      'lynx', 'chimera'],
-          # choices for the list widget in the GUI
-          label => 'WWW browser program'
-          # label for the GUI instead of 'browser'
-          ],
-         ['foo',
-          '=f',    # option accepts float value
-          undef,   # no default value
-          help => 'This is a short help',
-          # help string for usage() and the help balloon
-          longhelp => 'And this is a slightly longer help'
-          # longer help displayed in the GUI's help window
-          ]);
+        ('First section',
+	 ['', '', 'Section description'],
+	 ['debug', '!',  0],
+         ['age',   '=i', 18],
+
+	 'Second section',
+	 ['', '', 'Description for 2nd section'],
+         ['browser', '=s', 'tkweb'],
+         ['foo',     '=f', undef],
+        );
     new Tk::Getopt(-opttable => \@opttable,
                    -options => \%options);
 
@@ -1072,8 +1124,8 @@ two alternatives:
 
 =item *
 
-Use the B<-opttable> variant of C<new> and mark all non-GUI options with
-B<nogui>, e.g.
+Use the B<-opttable> variant of the C<new> constructor and mark all
+non-GUI options with B<nogui>, e.g.
 
     new Tk::Getopt(-opttable => ['not-in-gui', '!', undef,
                                  nogui => 1], ...)
@@ -1130,11 +1182,30 @@ a reference to the option editor window.
 
 Disable saving of options.
 
+=item -buttons
+
+Specify, which buttons should be drawn. It is advisable to draw at
+least the OK and Cancel buttons. The default set looks like this:
+
+    -buttons => [qw/ok apply cancel undo lastsaved save defaults/]
+
 =item -toplevel
 
-Use another widget instead of B<Toplevel> for embedding the option
-editor, e.g. C<Frame> to embed the editor into another toplevel widget (do not
-forget to pack the frame!).
+Use another widget class instead of B<Toplevel> for embedding the
+option editor, e.g. C<Frame> to embed the editor into another toplevel
+widget (do not forget to pack the frame!). See also the C<-pack>
+option below.
+
+=item -pack
+
+If using C<-toplevel> with a non-Toplevel widget (e.g. Frame) and
+using the C<-wait> option, then packing have to be done through the
+C<-pack> option. The argument to this option is a array reference of
+pack options, e.g.
+
+    $opt->option_editor(-toplevel => "Frame",
+                        -wait => 1,
+                        -pack => [-fill => "both", -expand => 1]);
 
 =item -statusbar
 
@@ -1144,7 +1215,7 @@ Use an additional status bar for help messages.
 
 Change button labels and title. This argument should be a hash reference
 with following keys: C<optedit>, C<undo>, C<lastsaved>, C<save>, C<defaults>,
-C<ok>, C<cancel>.
+C<ok>, C<cancel>, C<helpfor>.
 
 =item -wait
 
@@ -1163,7 +1234,7 @@ Note: this method returns immediately to the calling program.
 
 Buttons in the option editor window:
 
-=over 8
+=over 4
 
 =item OK
 
@@ -1191,7 +1262,7 @@ in C<new>.
 
 The option types are translated to following widgets:
 
-=over 8
+=over 4
 
 =item Boolean
 
@@ -1211,6 +1282,106 @@ B<FileDialog> if B<subtype> is set to B<file>.
 
 =back
 
+=head1 OPTTABLE ARGUMENTS
+
+Additional attributes in an option description have to be key-value
+pairs with following keys:
+
+=over
+
+=item alias
+
+An array of aliases also accepted by F<Getopt::Long>.
+
+=item label
+
+A label to be displayed in the GUI instead of the option name.
+
+=item help
+
+A short help string used by B<usage> and the Balloon help facility in
+B<option_editor>.
+
+=item longhelp
+
+A long help string used by B<option_editor>.
+
+=item choices
+
+An array of additional choices for the option editor.
+
+=item range
+
+An array with the beginning and end of a range for an integer or float value.
+
+=item strict
+
+Must be used with B<choices> or B<range>. When set to true, options have to
+match either the choices or the range.
+
+=item subtype
+
+The only permitted subtypes are I<file> and I<dir> to be used with
+string options. The GUI interface will pop up a file dialog for this
+option (either for selecting files or directories).
+
+=item var
+
+Use variable instead of I<$options-E<gt>{optname}> or I<$opt_optname>
+to store the value.
+
+=item nogui
+
+This option will not have an entry in the GUI.
+
+=item widget
+
+This should be a reference to a subroutine for creating an own widget.
+Folowing arguments will be passed to this subroutine:
+a reference to the F<Tk::Getopt> object, Frame object, options entry.
+The subroutine should create a widget in the frame (packing is not
+necessary!) and should return a reference to the created widget.
+
+=back
+
+Here is an example for using a complex opttable description:
+
+    @opttable =
+        ('Misc',   # Head of first group
+	 ['debug', # name of the option (--debug)
+          '!',     # type boolean, accept --nodebug
+          0,       # default is 0 (false)
+          callback => sub { $^W = 1
+                                if $options->{'debug'}; }
+          # additional attribute: callback to be called if
+          # you set or change the value
+          ],
+         ['age',
+          '=i',    # option accepts integer value
+          18,
+          strict => 1, # value must be in range
+          range => [0, 100], # allowed range
+          alias => ['year', 'years'] # possible aliases
+          ],
+	 'External', # Head of second group
+         ['browser',
+          '=s',    # option accepts string value
+          'tkweb',
+          choices => ['mosaic', 'netscape',
+                      'lynx', 'chimera'],
+          # choices for the list widget in the GUI
+          label => 'WWW browser program'
+          # label for the GUI instead of 'browser'
+          ],
+         ['foo',
+          '=f',    # option accepts float value
+          undef,   # no default value
+          help => 'This is a short help',
+          # help string for usage() and the help balloon
+          longhelp => 'And this is a slightly longer help'
+          # longer help displayed in the GUI's help window
+          ]);
+
 =head1 REQUIREMENTS
 
 You need at least:
